/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * $URL: https://scummvm.svn.sourceforge.net/svnroot/scummvm/scummvm/trunk/common/file.h $
 * $Id: file.h 33412 2008-07-29 16:09:10Z fingolfin $
 *
 */

#include "common/archives.h"
#include "common/fs.h"
#include "common/file.h"
#include "common/util.h"

namespace Common {


int Archive::getAllNames(StringList &list) {
	return matchPattern(list, "*");
}


FSDirectory::FSDirectory() : Archive(), _node(), _cached(false), _depth(1), _unbound(true) {

}

FSDirectory::FSDirectory(const FilesystemNode &node, int depth) : Archive(), _node(), _cached(false), _depth(depth), _unbound(false) {
	if (node.isDirectory()) {
		_node = node;
	}
}

FSDirectory::FSDirectory(const Common::String &name, int depth) : Archive(), _node(), _cached(false), _depth(depth), _unbound(true) {
	FilesystemNode node(name);
	if (node.isDirectory()) {
		_node = node;
		_unbound = false;
	}
}

FSDirectory::~FSDirectory() {

}

String FSDirectory::getName() const {
	if (_unbound) {
		// I hope this is the best thing to do if the directory is unbound.
		return String::emptyString;
	}

	return _node.getName();
}

FilesystemNode FSDirectory::lookupCache(NodeCache &cache, const Common::String &name) {
	// make caching as lazy as possible
	if (name.empty()) {
		return FilesystemNode();
	}

	cacheDirectory();

	if (cache.empty()) {
		return FilesystemNode();
	}

	FilesystemNode match;

	if (cache.contains(name)) {
		match = cache[name];
	}

	return match;
}

bool FSDirectory::hasFile(const String &name) {
	if (_unbound || name.empty()) {
		return false;
	}

	// Ensure name is not a pattern.
	if (name.contains("?*")) {
		warning("FSDirectory::hasFile: Can't be given a pattern to look for");
		return false;
	}

	FilesystemNode node = lookupCache(_fileCache, name);
	return node.exists();
}

SharedPtr<SeekableReadStream> FSDirectory::openFile(const Common::String &name) {
	if (_unbound || name.empty()) {
		return SharedPtr<SeekableReadStream>();
	}

	// Ensure name is not a pattern.
	if (name.contains("?*")) {
		warning("FSDirectory::openFile: Can't be given a pattern to look for");
		return SharedPtr<SeekableReadStream>();
	}

	FilesystemNode node = lookupCache(_fileCache, name);

	if (!node.exists()) {
		warning("FSDirectory::openFile: Trying to open a FilesystemNode which does not exist");
		return SharedPtr<SeekableReadStream>();
	} else if (node.isDirectory()) {
		warning("FSDirectory::openFile: Trying to open a FilesystemNode which is a directory");
		return SharedPtr<SeekableReadStream>();
	}

	SeekableReadStream *stream = node.openForReading();
	if (!stream) {
		warning("FSDirectory::openFile: Can't create stream for file '%s'", name.c_str());
	}

	return SharedPtr<SeekableReadStream>(stream);
}

Common::SharedPtr<FSDirectory> FSDirectory::getSubDirectory(const Common::String &name) {
	if (_unbound || name.empty()) {
		// NOTE: this is pretty funny, you could go on creating subdirs of an unbound directory,
		// and the only way out is checking if the getName() members returns an empty string...
		return Common::SharedPtr<FSDirectory>();
	}

	FilesystemNode node = lookupCache(_subDirCache, name);
	return Common::SharedPtr<FSDirectory>(new FSDirectory(node));
}

void FSDirectory::cacheDirectory() {
	if (_cached) {
		return;
	}

	cacheDirectoryRecursive(_node, _depth, "");

	_cached = true;
}

void FSDirectory::cacheDirectoryRecursive(FilesystemNode node, int depth, const String& prefix) {
	if (depth <= 0) {
		return;
	}

	FSList list;
	node.getChildren(list, FilesystemNode::kListAll, false);

	FSList::iterator it = list.begin();
	for ( ; it != list.end(); it++) {
		Common::String name = prefix + (*it).getName();

		// since the hashmap will be case insensitive, we need to check
		// for clashes when caching
		if ((*it).isDirectory()) {
			if (_subDirCache.contains(name)) {
				warning("FSDirectory::cacheDirectory: name clash when building cache, ignoring sub-directory '%s'", name.c_str());
			} else {
				cacheDirectoryRecursive(*it, depth - 1, name + "/");
				_subDirCache[name] = *it;
			}
		} else {
			if (_fileCache.contains(name)) {
				warning("FSDirectory::cacheDirectory: name clash when building cache, ignoring file '%s'", name.c_str());
			} else {
				_fileCache[name] = *it;
			}
		}
	}

}

int FSDirectory::matchPattern(StringList &list, const String &pattern) {
	if (_unbound || pattern.empty()) {
		return 0;
	}

	int matches = 0;

	// NOTE: what do we match pattern against? It is clear that we require
	// the pattern to contain at least the filename (with wildcards). How
	// do we match optional path elements in the pattern? We can do:
	//
	// * full path match: we require filename and each subdir's name to
	//   match. I.e. we don't accept a match if a pattern is shorter than
	//   the full path stored in the key.
	//
	// * partial path match: we require only the filename and the path
	//   elements specified in the pattern to match.

	// TODO: implement matching!

	return matches;
}


SearchSet::SearchSet() {

}

SearchSet::~SearchSet() {

}

void SearchSet::add(const String& name, SharedPtr<Archive> archive) {
	// NOTE: right now we blindly replace an archive if it is already present.
	// print out a warning in this case?
	_list.setVal(name, archive);
}

void SearchSet::remove(const String& name) {
	// NOTE: print out a warning in case the archive wasn't present?
	_list.erase(name);
}

void SearchSet::clear() {
	_list.clear();
}

// TODO: would it be better to scan all the set for clashes?
SharedPtr<SeekableReadStream> SearchSet::openFile(const String &name) {
	if (name.empty()) {
		return SharedPtr<SeekableReadStream>();
	}

	ArchiveList::iterator it = _list.begin();
	for ( ; it != _list.end(); it++) {
		if ((*it)._value->hasFile(name)) {
			return (*it)._value->openFile(name);
		}
	}

	return SharedPtr<SeekableReadStream>();
}

bool SearchSet::hasFile(const String &name) {
	if (name.empty()) {
		return false;
	}

	ArchiveList::iterator it = _list.begin();
	for ( ; it != _list.end(); it++) {
		if ((*it)._value->hasFile(name)) {
			return true;
		}
	}

	return false;
}

int SearchSet::matchPattern(StringList &list, const String &pattern) {
	// Shall we short circuit out if pattern is empty?

	int matches = 0;

	ArchiveList::iterator it = _list.begin();
	for ( ; it != _list.end(); it++) {
		matches += (*it)._value->matchPattern(list, pattern);
	}

	return matches;
}


} // namespace Common
