/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * $URL: https://scummvm.svn.sourceforge.net/svnroot/scummvm/scummvm/trunk/common/file.h $
 * $Id: file.h 33412 2008-07-29 16:09:10Z fingolfin $
 *
 */

#ifndef COMMON_ARCHIVES_H
#define COMMON_ARCHIVES_H

#include "common/fs.h"
#include "common/str.h"
#include "common/hash-str.h"
#include "common/list.h"
#include "common/ptr.h"
#include "common/stream.h"

namespace Common {


/*
	Archive defines an interface through which a set of files can be accessed.
*/
class Archive {
public:
	Archive() { }
	virtual ~Archive() { }

	// checks for existence
	virtual bool hasFile(const String &name) = 0;

	// returns a stream bound to a file in the archive
	virtual SharedPtr<SeekableReadStream> openFile(const String &name) = 0;

	// adds names to the list according to the specified pattern (they can
	// be used as parameters in fileOpen).
	// Returns the number of items added to the list (useful if the list is
	// not empty in the beginning)
	virtual int matchPattern(StringList &list, const String &pattern) = 0;

	// adds all the names in the archive.
	// Returns the number of items added to the list (useful if the list is
	// not empty in the beginning)
	virtual int getAllNames(StringList &list);
};

/*
	FSDirectory models a directory tree from the filesystem and allows users
	to access it through the Archive interface. It currently caches the
	underlying FilesystemNodes to speed up retrieval.

	FSDirectory can represent a single directory, or a tree with specified
	depth, rooted in a 'base' directory.
*/
class FSDirectory : public Archive {
	FilesystemNode	_node;

	// Caches are case insensitive, clashes are dealt with when creating
	// Original case is retained into key, though. Could it be used for
	// explanatory error messages/warning?
	typedef HashMap<String, FilesystemNode, IgnoreCase_Hash, IgnoreCase_EqualTo> NodeCache;
	NodeCache	_fileCache, _subDirCache;

	// look for a match
	FilesystemNode lookupCache(NodeCache &cache, const String &name);

	// cache management
	void cacheDirectory();
	void cacheDirectoryRecursive(FilesystemNode node, int depth, const String& prefix);
	bool _cached;
	int	_depth;

	// NOTE: FilesystemNode::getName(), as well as others informative methods, assert out
	// when the node is pathless. Sadly, we have no means to detect if a constructed node
	// is valid or not. The _unbound field is used to mark instances not linked to a real
	// directory.
	bool _unbound;

public:
	FSDirectory();
	FSDirectory(const String &name, int depth = 1);
	FSDirectory(const FilesystemNode &node, int depth = 1);
	~FSDirectory();

	String getName() const;

	// return a stream bound to a file in the current directory
	SharedPtr<SeekableReadStream> openFile(const String &name);

	// check for existence of the specified file
	bool hasFile(const String &name);

	int matchPattern(StringList &list, const String &pattern);

	// open a sub of the current directory
	Common::SharedPtr<FSDirectory> getSubDirectory(const String &name);
};


/*
	SearchSet models a set of Archives, thus allowing easy file retrieval
	from a moltitude of sources. Its intended usage is a situation in which
	there are no name clashes between files across the whole set.
*/
class SearchSet : public Archive {
	typedef SharedPtr<Archive> ArchivePtr;
	typedef HashMap<String, ArchivePtr, IgnoreCase_Hash, IgnoreCase_EqualTo> ArchiveList;
	ArchiveList _list;

public:
	SearchSet();
	~SearchSet();

	// adds a new Archive to the searchable set
	void add(const String& name, SharedPtr<Archive> archive);
	// removes an Archive from the set
	void remove(const String& name);
	// clears the set
	void clear();


	// currently open the first file whose name matches
	SharedPtr<SeekableReadStream> openFile(const String &name);
	bool hasFile(const String &name);
	int matchPattern(StringList &list, const String &pattern);
};


} // namespace Common

#endif
